@@ -752,38 +752,6 @@ BOOL freerdp_settings_set_default_order_support(rdpSettings* settings)
 	return TRUE;
 }
 
-BOOL freerdp_capability_buffer_allocate(rdpSettings* settings, UINT32 count)
-{
-	WINPR_ASSERT(settings);
-	WINPR_ASSERT(count > 0);
-	WINPR_ASSERT(count == 32);
-
-	freerdp_capability_buffer_free(settings);
-	WINPR_ASSERT(settings->ReceivedCapabilitiesSize == 0);
-
-	settings->ReceivedCapabilitiesSize = count;
-	void* tmp = realloc(settings->ReceivedCapabilities, count * sizeof(BYTE));
-	if (!tmp)
-		return FALSE;
-	memset(tmp, 0, count * sizeof(BYTE));
-	settings->ReceivedCapabilities = tmp;
-
-	tmp = realloc((void*)settings->ReceivedCapabilityData, count * sizeof(BYTE*));
-	if (!tmp)
-		return FALSE;
-	memset(tmp, 0, count * sizeof(BYTE*));
-	settings->ReceivedCapabilityData = (BYTE**)tmp;
-
-	tmp = realloc(settings->ReceivedCapabilityDataSizes, count * sizeof(UINT32));
-	if (!tmp)
-		return FALSE;
-	memset(tmp, 0, count * sizeof(UINT32));
-	settings->ReceivedCapabilityDataSizes = tmp;
-
-	return (settings->ReceivedCapabilities && settings->ReceivedCapabilityData &&
-	        settings->ReceivedCapabilityDataSizes);
-}
-
 #if !defined(WITH_FULL_CONFIG_PATH)
 static char* freerdp_settings_get_legacy_config_path(void)
 {
@@ -827,6 +795,19 @@ rdpSettings* freerdp_settings_new(DWORD flags)
 
 	if (!server && !remote)
 	{
+		if (!freerdp_settings_set_string(settings, FreeRDP_GatewayAvdScope,
+		                                 "https%%3A%%2F%%2F%s%%2F%s%%2Foauth2%%2Fnativeclient"))
+			goto out_fail;
+		if (!freerdp_settings_set_string(settings, FreeRDP_GatewayAvdAccessTokenFormat,
+		                                 "ms-appx-web%%3a%%2f%%2fMicrosoft.AAD.BrokerPlugin%%2f%s"))
+			goto out_fail;
+		if (!freerdp_settings_set_string(settings, FreeRDP_GatewayAvdAccessAadFormat,
+		                                 "https%%3A%%2F%%2F%s%%2F%s%%2Foauth2%%2Fnativeclient"))
+			goto out_fail;
+		if (!freerdp_settings_set_string(settings, FreeRDP_GatewayAvdScope,
+		                                 "https%3A%2F%2Fwww.wvd.microsoft.com%2F.default"))
+
+			goto out_fail;
 		if (!freerdp_settings_set_string(settings, FreeRDP_GatewayAvdClientID,
 		                                 "a85cf173-4192-42f8-81fa-777a763e6e2c"))
 			goto out_fail;
@@ -916,6 +897,7 @@ rdpSettings* freerdp_settings_new(DWORD flags)
 	    !freerdp_settings_set_bool(settings, FreeRDP_RdstlsSecurity, FALSE) ||
 	    !freerdp_settings_set_bool(settings, FreeRDP_NegotiateSecurityLayer, TRUE) ||
 	    !freerdp_settings_set_bool(settings, FreeRDP_RestrictedAdminModeRequired, FALSE) ||
+	    !freerdp_settings_set_bool(settings, FreeRDP_RestrictedAdminModeSupported, TRUE) ||
 	    !freerdp_settings_set_bool(settings, FreeRDP_MstscCookieMode, FALSE) ||
 	    !freerdp_settings_set_uint32(settings, FreeRDP_CookieMaxLength,
 	                                 DEFAULT_COOKIE_MAX_LENGTH) ||
@@ -998,7 +980,7 @@ rdpSettings* freerdp_settings_new(DWORD flags)
 	if (!freerdp_settings_set_pointer_len(settings, FreeRDP_RdpServerCertificate, NULL, 1))
 		goto out_fail;
 
-	if (!freerdp_capability_buffer_allocate(settings, 32))
+	if (!freerdp_capability_buffer_resize(settings, 32, FALSE))
 		goto out_fail;
 
 	{
@@ -1168,7 +1150,8 @@ rdpSettings* freerdp_settings_new(DWORD flags)
 	    !freerdp_settings_set_bool(settings, FreeRDP_GfxPlanar, TRUE) ||
 	    !freerdp_settings_set_bool(settings, FreeRDP_GfxH264, FALSE) ||
 	    !freerdp_settings_set_bool(settings, FreeRDP_GfxAVC444, FALSE) ||
-	    !freerdp_settings_set_bool(settings, FreeRDP_GfxSendQoeAck, FALSE))
+	    !freerdp_settings_set_bool(settings, FreeRDP_GfxSendQoeAck, FALSE) ||
+	    !freerdp_settings_set_bool(settings, FreeRDP_SupportDisplayControl, TRUE))
 		goto out_fail;
 	{
 		ARC_CS_PRIVATE_PACKET cookie = { 0 };
@@ -1355,7 +1338,7 @@ static BOOL freerdp_settings_int_buffer_copy(rdpSettings* _settings, const rdpSe
 
 	if (!freerdp_server_license_issuers_copy(_settings, settings->ServerLicenseProductIssuers,
 	                                         settings->ServerLicenseProductIssuersCount))
-		return FALSE;
+		goto out_fail;
 
 	if (settings->RdpServerCertificate)
 	{
@@ -1389,13 +1372,17 @@ static BOOL freerdp_settings_int_buffer_copy(rdpSettings* _settings, const rdpSe
 	                                 freerdp_settings_get_uint32(settings, FreeRDP_ChannelCount)))
 		goto out_fail;
 
-	_settings->OrderSupport = malloc(32);
-	if (!_settings->OrderSupport)
-		goto out_fail;
+	if (settings->OrderSupport)
+	{
+		_settings->OrderSupport = calloc(32, sizeof(BYTE));
+		if (!_settings->OrderSupport)
+			goto out_fail;
+
+		CopyMemory(_settings->OrderSupport, settings->OrderSupport, 32);
+	}
 
 	if (!freerdp_capability_buffer_copy(_settings, settings))
 		goto out_fail;
-	CopyMemory(_settings->OrderSupport, settings->OrderSupport, 32);
 
 	const UINT32 glyphCacheCount = 10;
 	const GLYPH_CACHE_DEFINITION* glyphCache =
@@ -1426,7 +1413,7 @@ static BOOL freerdp_settings_int_buffer_copy(rdpSettings* _settings, const rdpSe
 
 	const UINT32 nrports = freerdp_settings_get_uint32(settings, FreeRDP_TargetNetAddressCount);
 	if (!freerdp_target_net_adresses_reset(_settings, nrports))
-		return FALSE;
+		goto out_fail;
 
 	for (UINT32 i = 0; i < nrports; i++)
 	{
@@ -1434,14 +1421,17 @@ static BOOL freerdp_settings_int_buffer_copy(rdpSettings* _settings, const rdpSe
 		    freerdp_settings_get_pointer_array(settings, FreeRDP_TargetNetAddresses, i);
 		const UINT32* port =
 		    freerdp_settings_get_pointer_array(settings, FreeRDP_TargetNetPorts, i);
+		WINPR_ASSERT(address);
+		WINPR_ASSERT(port);
+
 		if (!freerdp_settings_set_pointer_array(_settings, FreeRDP_TargetNetAddresses, i, address))
-			return FALSE;
+			goto out_fail;
 		if (!freerdp_settings_set_pointer_array(_settings, FreeRDP_TargetNetPorts, i, port))
-			return FALSE;
+			goto out_fail;
 	}
 
 	{
-		const UINT32 len = freerdp_settings_get_uint32(_settings, FreeRDP_DeviceArraySize);
+		const UINT32 len = freerdp_settings_get_uint32(settings, FreeRDP_DeviceArraySize);
 		const UINT32 count = freerdp_settings_get_uint32(settings, FreeRDP_DeviceCount);
 
 		if (len < count)
@@ -1455,9 +1445,12 @@ static BOOL freerdp_settings_int_buffer_copy(rdpSettings* _settings, const rdpSe
 		{
 			const RDPDR_DEVICE* argv =
 			    freerdp_settings_get_pointer_array(settings, FreeRDP_DeviceArray, index);
+			WINPR_ASSERT(argv);
 			if (!freerdp_settings_set_pointer_array(_settings, FreeRDP_DeviceArray, index, argv))
 				goto out_fail;
 		}
+		WINPR_ASSERT(_settings->DeviceArray || (len == 0));
+		WINPR_ASSERT(len >= count);
 	}
 	{
 		const UINT32 len = freerdp_settings_get_uint32(_settings, FreeRDP_StaticChannelArraySize);
@@ -1474,6 +1467,7 @@ static BOOL freerdp_settings_int_buffer_copy(rdpSettings* _settings, const rdpSe
 		{
 			const ADDIN_ARGV* argv =
 			    freerdp_settings_get_pointer_array(settings, FreeRDP_StaticChannelArray, index);
+			WINPR_ASSERT(argv);
 			if (!freerdp_settings_set_pointer_array(_settings, FreeRDP_StaticChannelArray, index,
 			                                        argv))
 				goto out_fail;
@@ -1494,6 +1488,7 @@ static BOOL freerdp_settings_int_buffer_copy(rdpSettings* _settings, const rdpSe
 		{
 			const ADDIN_ARGV* argv =
 			    freerdp_settings_get_pointer_array(settings, FreeRDP_DynamicChannelArray, index);
+			WINPR_ASSERT(argv);
 			if (!freerdp_settings_set_pointer_array(_settings, FreeRDP_DynamicChannelArray, index,
 			                                        argv))
 				goto out_fail;
@@ -1520,34 +1515,6 @@ BOOL freerdp_settings_copy(rdpSettings* _settings, const rdpSettings* settings)
 	 * NULL to fix issues during cleanup */
 	rc = freerdp_settings_clone_keys(_settings, settings);
 
-	_settings->LoadBalanceInfo = NULL;
-	_settings->ServerRandom = NULL;
-	_settings->ClientRandom = NULL;
-	_settings->ServerCertificate = NULL;
-	_settings->RdpServerCertificate = NULL;
-	_settings->RdpServerRsaKey = NULL;
-	_settings->ChannelDefArray = NULL;
-	_settings->MonitorDefArray = NULL;
-	_settings->MonitorIds = NULL;
-	_settings->OrderSupport = NULL;
-	_settings->BitmapCacheV2CellInfo = NULL;
-	_settings->GlyphCache = NULL;
-	_settings->FragCache = NULL;
-	_settings->ClientAutoReconnectCookie = NULL;
-	_settings->ServerAutoReconnectCookie = NULL;
-	_settings->ClientTimeZone = NULL;
-	_settings->RedirectionPassword = NULL;
-	_settings->RedirectionTsvUrl = NULL;
-	_settings->TargetNetAddresses = NULL;
-	_settings->TargetNetPorts = NULL;
-	_settings->RedirectionGuid = NULL;
-	_settings->DeviceArray = NULL;
-	_settings->StaticChannelArray = NULL;
-	_settings->DynamicChannelArray = NULL;
-	_settings->ReceivedCapabilities = NULL;
-	_settings->ReceivedCapabilityData = NULL;
-	_settings->ReceivedCapabilityDataSizes = NULL;
-
 	_settings->ServerLicenseProductIssuersCount = 0;
 	_settings->ServerLicenseProductIssuers = NULL;
 
@@ -1635,8 +1602,9 @@ BOOL identity_set_from_smartcard_hash(SEC_WINNT_AUTH_IDENTITY_W* identity,
                                       const rdpSettings* settings,
                                       FreeRDP_Settings_Keys_String userId,
                                       FreeRDP_Settings_Keys_String domainId,
-                                      FreeRDP_Settings_Keys_String pwdId, const BYTE* certSha1,
-                                      size_t sha1len)
+                                      FreeRDP_Settings_Keys_String pwdId,
+                                      WINPR_ATTR_UNUSED const BYTE* certSha1,
+                                      WINPR_ATTR_UNUSED size_t sha1len)
 {
 #ifdef _WIN32
 	CERT_CREDENTIAL_INFO certInfo = { sizeof(CERT_CREDENTIAL_INFO), { 0 } };
@@ -1695,22 +1663,11 @@ BOOL freerdp_target_net_adresses_reset(rdpSettings* settings, size_t size)
 {
 	freerdp_target_net_addresses_free(settings);
 
-	if (size > 0)
-	{
-		if (!freerdp_settings_set_pointer_len_(settings, FreeRDP_TargetNetPorts,
-		                                       FreeRDP_UINT32_UNUSED, NULL, size, sizeof(UINT32)))
-			return FALSE;
-		if (!freerdp_settings_set_pointer_len_(settings, FreeRDP_TargetNetAddresses,
-		                                       FreeRDP_TargetNetAddressCount, NULL, size,
-		                                       sizeof(char*)))
-			return FALSE;
-	}
-	return TRUE;
+	return freerdp_target_net_addresses_resize(settings, size);
 }
 
 BOOL freerdp_settings_enforce_monitor_exists(rdpSettings* settings)
 {
-	const UINT32 nrIds = freerdp_settings_get_uint32(settings, FreeRDP_NumMonitorIds);
 	const UINT32 count = freerdp_settings_get_uint32(settings, FreeRDP_MonitorCount);
 	const BOOL fullscreen = freerdp_settings_get_bool(settings, FreeRDP_Fullscreen);
 	const BOOL multimon = freerdp_settings_get_bool(settings, FreeRDP_UseMultimon);
@@ -1723,11 +1680,6 @@ BOOL freerdp_settings_enforce_monitor_exists(rdpSettings* settings)
 			return FALSE;
 	}
 
-	if (nrIds == 0)
-	{
-		if (!freerdp_settings_set_uint32(settings, FreeRDP_NumMonitorIds, 1))
-			return FALSE;
-	}
 	if (!useMonitors || (count == 0))
 	{
 		const UINT32 width = freerdp_settings_get_uint32(settings, FreeRDP_DesktopWidth);
@@ -1762,3 +1714,19 @@ BOOL freerdp_settings_enforce_monitor_exists(rdpSettings* settings)
 
 	return TRUE;
 }
+
+BOOL freerdp_settings_enforce_consistency(rdpSettings* settings)
+{
+	if (freerdp_settings_get_bool(settings, FreeRDP_BitmapCacheV3Enabled))
+	{
+		settings->OrderSupportFlagsEx |= CACHE_BITMAP_V3_SUPPORT;
+		settings->OrderSupportFlags |= ORDER_FLAGS_EXTRA_SUPPORT;
+	}
+
+	if (settings->FrameMarkerCommandEnabled)
+	{
+		settings->OrderSupportFlagsEx |= ALTSEC_FRAME_MARKER_SUPPORT;
+		settings->OrderSupportFlags |= ORDER_FLAGS_EXTRA_SUPPORT;
+	}
+	return TRUE;
+}
