@@ -29,6 +29,7 @@
 #include <ctype.h>
 
 #include <winpr/assert.h>
+#include <winpr/cast.h>
 #include <winpr/asn1.h>
 #include <winpr/crt.h>
 #include <winpr/interlocked.h>
@@ -146,8 +147,7 @@ static krb5_error_code kerberos_log_msg(krb5_context ctx, krb5_error_code code,
 			if (WLog_IsLevelActive(log, level))
 			{
 				const char* msg = krb5_get_error_message(ctx, code);
-				WLog_PrintMessage(log, WLOG_MESSAGE_TEXT, level, line, file, fkt, "%s (%s [%d])",
-				                  what, msg, code);
+				WLog_PrintTextMessage(log, level, line, file, fkt, "%s (%s [%d])", what, msg, code);
 				krb5_free_error_message(ctx, msg);
 			}
 		}
@@ -197,8 +197,10 @@ static KRB_CONTEXT* kerberos_ContextNew(KRB_CREDENTIALS* credentials)
 	return context;
 }
 
-static krb5_error_code krb5_prompter(krb5_context context, void* data, const char* name,
-                                     const char* banner, int num_prompts, krb5_prompt prompts[])
+static krb5_error_code krb5_prompter(krb5_context context, void* data,
+                                     WINPR_ATTR_UNUSED const char* name,
+                                     WINPR_ATTR_UNUSED const char* banner, int num_prompts,
+                                     krb5_prompt prompts[])
 {
 	for (int i = 0; i < num_prompts; i++)
 	{
@@ -206,7 +208,11 @@ static krb5_error_code krb5_prompter(krb5_context context, void* data, const cha
 		if (type && (type == KRB5_PROMPT_TYPE_PREAUTH || type == KRB5_PROMPT_TYPE_PASSWORD) && data)
 		{
 			prompts[i].reply->data = _strdup((const char*)data);
-			prompts[i].reply->length = strlen((const char*)data);
+
+			const size_t len = strlen((const char*)data);
+			if (len > UINT32_MAX)
+				return KRB5KRB_ERR_GENERIC;
+			prompts[i].reply->length = (UINT32)len;
 		}
 	}
 	return 0;
@@ -238,14 +244,42 @@ static BOOL isValidIP(const char* ipAddress)
 	return isValidIPv4(ipAddress) || isValidIPv6(ipAddress);
 }
 
-static int build_krbtgt(krb5_context ctx, krb5_data* realm, krb5_principal* ptarget)
+#if defined(WITH_KRB5_MIT)
+WINPR_ATTR_MALLOC(free, 1)
+static char* get_realm_name(krb5_data realm, size_t* plen)
+{
+	WINPR_ASSERT(plen);
+	*plen = 0;
+	if ((realm.length <= 0) || (!realm.data))
+		return NULL;
+
+	char* name = NULL;
+	(void)winpr_asprintf(&name, plen, "krbtgt/%*s@%*s", realm.length, realm.data, realm.length,
+	                     realm.data);
+	return name;
+}
+#elif defined(WITH_KRB5_HEIMDAL)
+WINPR_ATTR_MALLOC(free, 1)
+static char* get_realm_name(Realm realm, size_t* plen)
+{
+	WINPR_ASSERT(plen);
+	*plen = 0;
+	if (!realm)
+		return NULL;
+
+	char* name = NULL;
+	(void)winpr_asprintf(&name, plen, "krbtgt/%s@%s", realm, realm);
+	return name;
+}
+#endif
+
+static int build_krbtgt(krb5_context ctx, krb5_principal principal, krb5_principal* ptarget)
 {
 	/* "krbtgt/" + realm + "@" + realm */
 	size_t len = 0;
-	char* name = NULL;
 	krb5_error_code rv = KRB5_CC_NOMEM;
 
-	(void)winpr_asprintf(&name, &len, "krbtgt/%s@%s", realm->data, realm->data);
+	char* name = get_realm_name(principal->realm, &len);
 	if (!name || (len == 0))
 		goto fail;
 
@@ -260,9 +294,10 @@ fail:
 #endif /* WITH_KRB5 */
 
 static SECURITY_STATUS SEC_ENTRY kerberos_AcquireCredentialsHandleA(
-    SEC_CHAR* pszPrincipal, SEC_CHAR* pszPackage, ULONG fCredentialUse, void* pvLogonID,
-    void* pAuthData, SEC_GET_KEY_FN pGetKeyFn, void* pvGetKeyArgument, PCredHandle phCredential,
-    PTimeStamp ptsExpiry)
+    SEC_CHAR* pszPrincipal, WINPR_ATTR_UNUSED SEC_CHAR* pszPackage, ULONG fCredentialUse,
+    WINPR_ATTR_UNUSED void* pvLogonID, void* pAuthData, WINPR_ATTR_UNUSED SEC_GET_KEY_FN pGetKeyFn,
+    WINPR_ATTR_UNUSED void* pvGetKeyArgument, PCredHandle phCredential,
+    WINPR_ATTR_UNUSED PTimeStamp ptsExpiry)
 {
 #ifdef WITH_KRB5
 	SEC_WINPR_KERBEROS_SETTINGS* krb_settings = NULL;
@@ -329,6 +364,7 @@ static SECURITY_STATUS SEC_ENTRY kerberos_AcquireCredentialsHandleA(
 
 		if (rv)
 			goto cleanup;
+		WINPR_ASSERT(principal);
 	}
 
 	if (krb_settings && krb_settings->cache)
@@ -401,13 +437,14 @@ static SECURITY_STATUS SEC_ENTRY kerberos_AcquireCredentialsHandleA(
 	{
 		krb5_creds creds = { 0 };
 		krb5_creds matchCreds = { 0 };
-		int matchFlags = KRB5_TC_MATCH_TIMES;
+		krb5_flags matchFlags = KRB5_TC_MATCH_TIMES;
 
 		krb5_timeofday(ctx, &matchCreds.times.endtime);
 		matchCreds.times.endtime += 60;
 		matchCreds.client = principal;
 
-		if (krb_log_exec(build_krbtgt, ctx, &principal->realm, &matchCreds.server))
+		WINPR_ASSERT(principal);
+		if (krb_log_exec(build_krbtgt, ctx, principal, &matchCreds.server))
 			goto cleanup;
 
 		int rv = krb5_cc_retrieve_cred(ctx, ccache, matchFlags, &matchCreds, &creds);
@@ -544,9 +581,8 @@ static SECURITY_STATUS SEC_ENTRY kerberos_FreeCredentialsHandle(PCredHandle phCr
 #endif
 }
 
-static SECURITY_STATUS SEC_ENTRY kerberos_QueryCredentialsAttributesW(PCredHandle phCredential,
-                                                                      ULONG ulAttribute,
-                                                                      void* pBuffer)
+static SECURITY_STATUS SEC_ENTRY kerberos_QueryCredentialsAttributesW(
+    WINPR_ATTR_UNUSED PCredHandle phCredential, ULONG ulAttribute, WINPR_ATTR_UNUSED void* pBuffer)
 {
 #ifdef WITH_KRB5
 	switch (ulAttribute)
@@ -659,116 +695,198 @@ cleanup:
 	return ret;
 }
 
+static BOOL append(char* dst, size_t dstSize, const char* src)
+{
+	const size_t dlen = strnlen(dst, dstSize);
+	const size_t slen = strlen(src);
+	if (dlen + slen >= dstSize)
+		return FALSE;
+	if (!strncat(dst, src, dstSize - dlen))
+		return FALSE;
+	return TRUE;
+}
+
+static BOOL kerberos_rd_tgt_req_tag2(WinPrAsn1Decoder* dec, char* buf, size_t len)
+{
+	BOOL rc = FALSE;
+	WinPrAsn1Decoder seq = { 0 };
+
+	/* server-name [2] PrincipalName (SEQUENCE) */
+	if (!WinPrAsn1DecReadSequence(dec, &seq))
+		goto end;
+
+	/* name-type [0] INTEGER */
+	BOOL error = FALSE;
+	WinPrAsn1_INTEGER val = 0;
+	if (!WinPrAsn1DecReadContextualInteger(&seq, 0, &error, &val))
+		goto end;
+
+	/* name-string [1] SEQUENCE OF GeneralString */
+	if (!WinPrAsn1DecReadContextualSequence(&seq, 1, &error, dec))
+		goto end;
+
+	WinPrAsn1_tag tag = 0;
+	BOOL first = TRUE;
+	while (WinPrAsn1DecPeekTag(dec, &tag))
+	{
+		BOOL success = FALSE;
+		char* lstr = NULL;
+		if (!WinPrAsn1DecReadGeneralString(dec, &lstr))
+			goto fail;
+
+		if (!first)
+		{
+			if (!append(buf, len, "/"))
+				goto fail;
+		}
+		first = FALSE;
+
+		if (!append(buf, len, lstr))
+			goto fail;
+
+		success = TRUE;
+	fail:
+		free(lstr);
+		if (!success)
+			goto end;
+	}
+
+	rc = TRUE;
+end:
+	return rc;
+}
+
+static BOOL kerberos_rd_tgt_req_tag3(WinPrAsn1Decoder* dec, char* buf, size_t len)
+{
+	/* realm [3] Realm */
+	BOOL rc = FALSE;
+	WinPrAsn1_STRING str = NULL;
+	if (!WinPrAsn1DecReadGeneralString(dec, &str))
+		goto end;
+
+	if (!append(buf, len, "@"))
+		goto end;
+	if (!append(buf, len, str))
+		goto end;
+
+	rc = TRUE;
+end:
+	free(str);
+	return rc;
+}
+
+static BOOL kerberos_rd_tgt_req(WinPrAsn1Decoder* dec, char** target)
+{
+	BOOL rc = FALSE;
+
+	if (!target)
+		return FALSE;
+	*target = NULL;
+
+	wStream s = WinPrAsn1DecGetStream(dec);
+	const size_t len = Stream_Length(&s);
+	if (len == 0)
+		return TRUE;
+
+	WinPrAsn1Decoder dec2 = { 0 };
+	WinPrAsn1_tagId tag = 0;
+	if (WinPrAsn1DecReadContextualTag(dec, &tag, &dec2) == 0)
+		return FALSE;
+
+	char* buf = calloc(len + 1, sizeof(char));
+	if (!buf)
+		return FALSE;
+
+	/* We expect ASN1 context tag values 2 or 3.
+	 *
+	 * In case we got value 2 an (optional) context tag value 3 might follow.
+	 */
+	BOOL checkForTag3 = TRUE;
+	if (tag == 2)
+	{
+		rc = kerberos_rd_tgt_req_tag2(&dec2, buf, len);
+		if (rc)
+		{
+			const size_t res = WinPrAsn1DecReadContextualTag(dec, &tag, dec);
+			if (res == 0)
+				checkForTag3 = FALSE;
+		}
+	}
+
+	if (checkForTag3)
+	{
+		if (tag == 3)
+			rc = kerberos_rd_tgt_req_tag3(&dec2, buf, len);
+		else
+			rc = FALSE;
+	}
+
+	if (rc)
+		*target = buf;
+	else
+		free(buf);
+	return rc;
+}
+
+static BOOL kerberos_rd_tgt_rep(WinPrAsn1Decoder* dec, krb5_data* ticket)
+{
+	if (!ticket)
+		return FALSE;
+
+	/* ticket [2] Ticket */
+	WinPrAsn1Decoder asnTicket = { 0 };
+	WinPrAsn1_tagId tag = 0;
+	if (WinPrAsn1DecReadContextualTag(dec, &tag, &asnTicket) == 0)
+		return FALSE;
+
+	if (tag != 2)
+		return FALSE;
+
+	wStream s = WinPrAsn1DecGetStream(&asnTicket);
+	ticket->data = Stream_BufferAs(&s, char);
+
+	const size_t len = Stream_Length(&s);
+	if (len > UINT32_MAX)
+		return FALSE;
+	ticket->length = (UINT32)len;
+	return TRUE;
+}
+
 static BOOL kerberos_rd_tgt_token(const sspi_gss_data* token, char** target, krb5_data* ticket)
 {
-	WinPrAsn1Decoder dec;
-	WinPrAsn1Decoder dec2;
 	BOOL error = 0;
-	WinPrAsn1_tagId tag = 0;
 	WinPrAsn1_INTEGER val = 0;
-	size_t len = 0;
-	wStream s;
-	char* buf = NULL;
-	char* str = NULL;
 
 	WINPR_ASSERT(token);
 
-	WinPrAsn1Decoder_InitMem(&dec, WINPR_ASN1_DER, (BYTE*)token->data, token->length);
+	if (target)
+		*target = NULL;
+
+	WinPrAsn1Decoder der = { 0 };
+	WinPrAsn1Decoder_InitMem(&der, WINPR_ASN1_DER, (BYTE*)token->data, token->length);
 
 	/* KERB-TGT-REQUEST (SEQUENCE) */
-	if (!WinPrAsn1DecReadSequence(&dec, &dec2))
+	WinPrAsn1Decoder seq = { 0 };
+	if (!WinPrAsn1DecReadSequence(&der, &seq))
 		return FALSE;
-	dec = dec2;
 
 	/* pvno [0] INTEGER */
-	if (!WinPrAsn1DecReadContextualInteger(&dec, 0, &error, &val) || val != 5)
+	if (!WinPrAsn1DecReadContextualInteger(&seq, 0, &error, &val) || val != 5)
 		return FALSE;
 
 	/* msg-type [1] INTEGER */
-	if (!WinPrAsn1DecReadContextualInteger(&dec, 1, &error, &val))
+	if (!WinPrAsn1DecReadContextualInteger(&seq, 1, &error, &val))
 		return FALSE;
 
-	if (val == KRB_TGT_REQ)
+	switch (val)
 	{
-		if (!target)
-			return FALSE;
-		*target = NULL;
-
-		s = WinPrAsn1DecGetStream(&dec);
-		len = Stream_Length(&s);
-		if (len == 0)
-			return TRUE;
-
-		buf = malloc(len);
-		if (!buf)
-			return FALSE;
-
-		*buf = 0;
-		*target = buf;
-
-		if (!WinPrAsn1DecReadContextualTag(&dec, &tag, &dec2))
-			goto fail;
-
-		if (tag == 2)
-		{
-			WinPrAsn1Decoder seq;
-			/* server-name [2] PrincipalName (SEQUENCE) */
-			if (!WinPrAsn1DecReadSequence(&dec2, &seq))
-				goto fail;
-
-			/* name-type [0] INTEGER */
-			if (!WinPrAsn1DecReadContextualInteger(&seq, 0, &error, &val))
-				goto fail;
-
-			/* name-string [1] SEQUENCE OF GeneralString */
-			if (!WinPrAsn1DecReadContextualSequence(&seq, 1, &error, &dec2))
-				goto fail;
-
-			while (WinPrAsn1DecPeekTag(&dec2, &tag))
-			{
-				if (!WinPrAsn1DecReadGeneralString(&dec2, &str))
-					goto fail;
-
-				if (buf != *target)
-					*buf++ = '/';
-				buf = stpcpy(buf, str);
-				free(str);
-			}
-
-			if (!WinPrAsn1DecReadContextualTag(&dec, &tag, &dec2))
-				return TRUE;
-		}
-
-		/* realm [3] Realm */
-		if (tag != 3 || !WinPrAsn1DecReadGeneralString(&dec2, &str))
-			goto fail;
-
-		*buf++ = '@';
-		strcpy(buf, str);
-		free(str);
-		return TRUE;
+		case KRB_TGT_REQ:
+			return kerberos_rd_tgt_req(&seq, target);
+		case KRB_TGT_REP:
+			return kerberos_rd_tgt_rep(&seq, ticket);
+		default:
+			break;
 	}
-	else if (val == KRB_TGT_REP)
-	{
-		if (!ticket)
-			return FALSE;
-
-		/* ticket [2] Ticket */
-		if (!WinPrAsn1DecReadContextualTag(&dec, &tag, &dec2) || tag != 2)
-			return FALSE;
-
-		s = WinPrAsn1DecGetStream(&dec2);
-		ticket->data = Stream_BufferAs(&s, char);
-		ticket->length = Stream_Length(&s);
-		return TRUE;
-	}
-	else
-		return FALSE;
-
-fail:
-	free(buf);
-	if (target)
-		*target = NULL;
 	return FALSE;
 }
 
@@ -778,11 +896,11 @@ static BOOL kerberos_hash_channel_bindings(WINPR_DIGEST_CTX* md5, SEC_CHANNEL_BI
 {
 	BYTE buf[4];
 
-	Data_Write_UINT32(buf, bindings->dwInitiatorAddrType);
+	winpr_Data_Write_UINT32(buf, bindings->dwInitiatorAddrType);
 	if (!winpr_Digest_Update(md5, buf, 4))
 		return FALSE;
 
-	Data_Write_UINT32(buf, bindings->cbInitiatorLength);
+	winpr_Data_Write_UINT32(buf, bindings->cbInitiatorLength);
 	if (!winpr_Digest_Update(md5, buf, 4))
 		return FALSE;
 
@@ -791,11 +909,11 @@ static BOOL kerberos_hash_channel_bindings(WINPR_DIGEST_CTX* md5, SEC_CHANNEL_BI
 	                         bindings->cbInitiatorLength))
 		return FALSE;
 
-	Data_Write_UINT32(buf, bindings->dwAcceptorAddrType);
+	winpr_Data_Write_UINT32(buf, bindings->dwAcceptorAddrType);
 	if (!winpr_Digest_Update(md5, buf, 4))
 		return FALSE;
 
-	Data_Write_UINT32(buf, bindings->cbAcceptorLength);
+	winpr_Data_Write_UINT32(buf, bindings->cbAcceptorLength);
 	if (!winpr_Digest_Update(md5, buf, 4))
 		return FALSE;
 
@@ -804,7 +922,7 @@ static BOOL kerberos_hash_channel_bindings(WINPR_DIGEST_CTX* md5, SEC_CHANNEL_BI
 	                         bindings->cbAcceptorLength))
 		return FALSE;
 
-	Data_Write_UINT32(buf, bindings->cbApplicationDataLength);
+	winpr_Data_Write_UINT32(buf, bindings->cbApplicationDataLength);
 	if (!winpr_Digest_Update(md5, buf, 4))
 		return FALSE;
 
@@ -818,8 +936,9 @@ static BOOL kerberos_hash_channel_bindings(WINPR_DIGEST_CTX* md5, SEC_CHANNEL_BI
 
 static SECURITY_STATUS SEC_ENTRY kerberos_InitializeSecurityContextA(
     PCredHandle phCredential, PCtxtHandle phContext, SEC_CHAR* pszTargetName, ULONG fContextReq,
-    ULONG Reserved1, ULONG TargetDataRep, PSecBufferDesc pInput, ULONG Reserved2,
-    PCtxtHandle phNewContext, PSecBufferDesc pOutput, ULONG* pfContextAttr, PTimeStamp ptsExpiry)
+    WINPR_ATTR_UNUSED ULONG Reserved1, WINPR_ATTR_UNUSED ULONG TargetDataRep, PSecBufferDesc pInput,
+    WINPR_ATTR_UNUSED ULONG Reserved2, PCtxtHandle phNewContext, PSecBufferDesc pOutput,
+    WINPR_ATTR_UNUSED ULONG* pfContextAttr, WINPR_ATTR_UNUSED PTimeStamp ptsExpiry)
 {
 #ifdef WITH_KRB5
 	PSecBuffer input_buffer = NULL;
@@ -990,8 +1109,8 @@ static SECURITY_STATUS SEC_ENTRY kerberos_InitializeSecurityContextA(
 			/* Write the checksum (delegation not implemented) */
 			cksum.data = cksum_contents;
 			cksum.length = sizeof(cksum_contents);
-			Data_Write_UINT32(cksum_contents, 16);
-			Data_Write_UINT32((cksum_contents + 20), context->flags);
+			winpr_Data_Write_UINT32(cksum_contents, 16);
+			winpr_Data_Write_UINT32((cksum_contents + 20), context->flags);
 
 			if (bindings_buffer)
 			{
@@ -1161,10 +1280,113 @@ static SECURITY_STATUS SEC_ENTRY kerberos_InitializeSecurityContextW(
 	return status;
 }
 
+#ifdef WITH_KRB5
+static BOOL retrieveTgtForPrincipal(KRB_CREDENTIALS* credentials, krb5_principal principal,
+                                    krb5_creds* creds)
+{
+	BOOL ret = FALSE;
+	krb5_kt_cursor cur = { 0 };
+	krb5_keytab_entry entry = { 0 };
+	if (krb_log_exec(krb5_kt_start_seq_get, credentials->ctx, credentials->keytab, &cur))
+		goto cleanup;
+
+	do
+	{
+		krb5_error_code rv =
+		    krb_log_exec(krb5_kt_next_entry, credentials->ctx, credentials->keytab, &entry, &cur);
+		if (rv == KRB5_KT_END)
+			break;
+		if (rv != 0)
+			goto cleanup;
+
+		if (krb5_principal_compare(credentials->ctx, principal, entry.principal))
+			break;
+		rv = krb_log_exec(krb5glue_free_keytab_entry_contents, credentials->ctx, &entry);
+		memset(&entry, 0, sizeof(entry));
+		if (rv)
+			goto cleanup;
+	} while (1);
+
+	if (krb_log_exec(krb5_kt_end_seq_get, credentials->ctx, credentials->keytab, &cur))
+		goto cleanup;
+
+	if (!entry.principal)
+		goto cleanup;
+
+	/* Get the TGT */
+	if (krb_log_exec(krb5_get_init_creds_keytab, credentials->ctx, creds, entry.principal,
+	                 credentials->keytab, 0, NULL, NULL))
+		goto cleanup;
+
+	ret = TRUE;
+
+cleanup:
+	return ret;
+}
+
+static BOOL retrieveSomeTgt(KRB_CREDENTIALS* credentials, const char* target, krb5_creds* creds)
+{
+	BOOL ret = TRUE;
+	krb5_principal target_princ = { 0 };
+	char* default_realm = NULL;
+
+	krb5_error_code rv =
+	    krb_log_exec(krb5_parse_name_flags, credentials->ctx, target, 0, &target_princ);
+	if (rv)
+		return FALSE;
+
+	if (!target_princ->realm.length)
+	{
+		rv = krb_log_exec(krb5_get_default_realm, credentials->ctx, &default_realm);
+		if (rv)
+			goto out;
+
+		target_princ->realm.data = default_realm;
+		target_princ->realm.length = (unsigned int)strlen(default_realm);
+	}
+
+	/*
+	 * First try with the account service. We were requested with something like
+	 * TERMSRV/<host>@<realm>, let's see if we have that in our keytab and if we're able
+	 * to retrieve a TGT with that entry
+	 *
+	 */
+	if (retrieveTgtForPrincipal(credentials, target_princ, creds))
+		goto out;
+
+	ret = FALSE;
+
+	/*
+	 * if it's not working let's try with <host>$@<REALM> (note the dollar)
+	 */
+	char hostDollar[300] = { 0 };
+	if (target_princ->length < 2)
+		goto out;
+
+	(void)snprintf(hostDollar, sizeof(hostDollar) - 1, "%s$@%s", target_princ->data[1].data,
+	               target_princ->realm.data);
+	krb5_free_principal(credentials->ctx, target_princ);
+
+	rv = krb_log_exec(krb5_parse_name_flags, credentials->ctx, hostDollar, 0, &target_princ);
+	if (rv)
+		return FALSE;
+
+	ret = retrieveTgtForPrincipal(credentials, target_princ, creds);
+
+out:
+	if (default_realm)
+		krb5_free_default_realm(credentials->ctx, default_realm);
+
+	krb5_free_principal(credentials->ctx, target_princ);
+	return ret;
+}
+#endif
+
 static SECURITY_STATUS SEC_ENTRY kerberos_AcceptSecurityContext(
-    PCredHandle phCredential, PCtxtHandle phContext, PSecBufferDesc pInput, ULONG fContextReq,
-    ULONG TargetDataRep, PCtxtHandle phNewContext, PSecBufferDesc pOutput, ULONG* pfContextAttr,
-    PTimeStamp ptsExpity)
+    PCredHandle phCredential, PCtxtHandle phContext, PSecBufferDesc pInput,
+    WINPR_ATTR_UNUSED ULONG fContextReq, WINPR_ATTR_UNUSED ULONG TargetDataRep,
+    PCtxtHandle phNewContext, PSecBufferDesc pOutput, ULONG* pfContextAttr,
+    WINPR_ATTR_UNUSED PTimeStamp ptsExpity)
 {
 #ifdef WITH_KRB5
 	BOOL isNewContext = FALSE;
@@ -1178,11 +1400,7 @@ static SECURITY_STATUS SEC_ENTRY kerberos_AcceptSecurityContext(
 	krb5_flags ap_flags = 0;
 	krb5glue_authenticator authenticator = NULL;
 	char* target = NULL;
-	char* sname = NULL;
-	char* realm = NULL;
-	krb5_kt_cursor cur = { 0 };
 	krb5_keytab_entry entry = { 0 };
-	krb5_principal principal = NULL;
 	krb5_creds creds = { 0 };
 
 	/* behave like windows SSPIs that don't want empty context */
@@ -1231,55 +1449,7 @@ static SECURITY_STATUS SEC_ENTRY kerberos_AcceptSecurityContext(
 		if (!kerberos_rd_tgt_token(&input_token, &target, NULL))
 			goto bad_token;
 
-		if (target)
-		{
-			if (*target != 0 && *target != '@')
-				sname = target;
-			realm = strchr(target, '@');
-			if (realm)
-				realm++;
-		}
-
-		if (krb_log_exec(krb5_parse_name_flags, credentials->ctx, sname ? sname : "",
-		                 KRB5_PRINCIPAL_PARSE_NO_REALM, &principal))
-			goto cleanup;
-
-		if (realm)
-		{
-			if (krb_log_exec(krb5glue_set_principal_realm, credentials->ctx, principal, realm))
-				goto cleanup;
-		}
-
-		if (krb_log_exec(krb5_kt_start_seq_get, credentials->ctx, credentials->keytab, &cur))
-			goto cleanup;
-
-		do
-		{
-			krb5_error_code rv = krb_log_exec(krb5_kt_next_entry, credentials->ctx,
-			                                  credentials->keytab, &entry, &cur);
-			if (rv == KRB5_KT_END)
-				break;
-			if (rv != 0)
-				goto cleanup;
-
-			if ((!sname || krb_log_exec(krb5_principal_compare_any_realm, credentials->ctx,
-			                            principal, entry.principal)) &&
-			    (!realm ||
-			     krb_log_exec(krb5_realm_compare, credentials->ctx, principal, entry.principal)))
-				break;
-			if (krb_log_exec(krb5glue_free_keytab_entry_contents, credentials->ctx, &entry))
-				goto cleanup;
-		} while (1);
-
-		if (krb_log_exec(krb5_kt_end_seq_get, credentials->ctx, credentials->keytab, &cur))
-			goto cleanup;
-
-		if (!entry.principal)
-			goto cleanup;
-
-		/* Get the TGT */
-		if (krb_log_exec(krb5_get_init_creds_keytab, credentials->ctx, &creds, entry.principal,
-		                 credentials->keytab, 0, NULL, NULL))
+		if (!retrieveSomeTgt(credentials, target, &creds))
 			goto cleanup;
 
 		if (!kerberos_mk_tgt_token(output_buffer, KRB_TGT_REP, NULL, NULL, &creds.ticket))
@@ -1511,7 +1681,7 @@ static SECURITY_STATUS kerberos_ATTR_TICKET_LOGON(KRB_CONTEXT* context,
 {
 	krb5_creds matchCred = { 0 };
 	krb5_auth_context authContext = NULL;
-	int getCredsFlags = KRB5_GC_CACHED;
+	krb5_flags getCredsFlags = KRB5_GC_CACHED;
 	BOOL firstRun = TRUE;
 	krb5_creds* hostCred = NULL;
 	SECURITY_STATUS ret = SEC_E_INSUFFICIENT_MEMORY;
@@ -1620,16 +1790,16 @@ static SECURITY_STATUS SEC_ENTRY kerberos_QueryContextAttributesW(PCtxtHandle ph
 	return kerberos_QueryContextAttributesA(phContext, ulAttribute, pBuffer);
 }
 
-static SECURITY_STATUS SEC_ENTRY kerberos_SetContextAttributesW(PCtxtHandle phContext,
-                                                                ULONG ulAttribute, void* pBuffer,
-                                                                ULONG cbBuffer)
+static SECURITY_STATUS SEC_ENTRY kerberos_SetContextAttributesW(
+    WINPR_ATTR_UNUSED PCtxtHandle phContext, WINPR_ATTR_UNUSED ULONG ulAttribute,
+    WINPR_ATTR_UNUSED void* pBuffer, WINPR_ATTR_UNUSED ULONG cbBuffer)
 {
 	return SEC_E_UNSUPPORTED_FUNCTION;
 }
 
-static SECURITY_STATUS SEC_ENTRY kerberos_SetContextAttributesA(PCtxtHandle phContext,
-                                                                ULONG ulAttribute, void* pBuffer,
-                                                                ULONG cbBuffer)
+static SECURITY_STATUS SEC_ENTRY kerberos_SetContextAttributesA(
+    WINPR_ATTR_UNUSED PCtxtHandle phContext, WINPR_ATTR_UNUSED ULONG ulAttribute,
+    WINPR_ATTR_UNUSED void* pBuffer, WINPR_ATTR_UNUSED ULONG cbBuffer)
 {
 	return SEC_E_UNSUPPORTED_FUNCTION;
 }
@@ -1637,7 +1807,7 @@ static SECURITY_STATUS SEC_ENTRY kerberos_SetContextAttributesA(PCtxtHandle phCo
 static SECURITY_STATUS SEC_ENTRY kerberos_SetCredentialsAttributesX(PCredHandle phCredential,
                                                                     ULONG ulAttribute,
                                                                     void* pBuffer, ULONG cbBuffer,
-                                                                    BOOL unicode)
+                                                                    WINPR_ATTR_UNUSED BOOL unicode)
 {
 #ifdef WITH_KRB5
 	KRB_CREDENTIALS* credentials = NULL;
@@ -1779,17 +1949,18 @@ static SECURITY_STATUS SEC_ENTRY kerberos_EncryptMessage(PCtxtHandle phContext,
 	encrypt_iov[1].data.data = data_buffer->pvBuffer;
 
 	/* Write the GSS header with 0 in RRC */
-	Data_Write_UINT16_BE(header, TOK_ID_WRAP);
-	header[2] = flags;
+	winpr_Data_Write_UINT16_BE(header, TOK_ID_WRAP);
+	header[2] = WINPR_ASSERTING_INT_CAST(char, flags);
 	header[3] = (char)0xFF;
-	Data_Write_UINT32(header + 4, 0);
-	Data_Write_UINT64_BE(header + 8, (context->local_seq + MessageSeqNo));
+	winpr_Data_Write_UINT32(header + 4, 0);
+	winpr_Data_Write_UINT64_BE(header + 8, (context->local_seq + MessageSeqNo));
 
 	/* Copy header to be encrypted */
 	CopyMemory(encrypt_iov[2].data.data, header, 16);
 
 	/* Set the correct RRC */
-	Data_Write_UINT16_BE(header + 6, 16 + encrypt_iov[3].data.length + encrypt_iov[4].data.length);
+	const size_t len = 16 + encrypt_iov[3].data.length + encrypt_iov[4].data.length;
+	winpr_Data_Write_UINT16_BE(header + 6, WINPR_ASSERTING_INT_CAST(UINT16, len));
 
 	if (krb_log_exec(krb5glue_encrypt_iov, creds->ctx, key, usage, encrypt_iov,
 	                 ARRAYSIZE(encrypt_iov)))
@@ -1811,7 +1982,6 @@ static SECURITY_STATUS SEC_ENTRY kerberos_DecryptMessage(PCtxtHandle phContext,
 	PSecBuffer data_buffer = NULL;
 	krb5glue_key key = NULL;
 	krb5_keyusage usage = 0;
-	char* header = NULL;
 	uint16_t tok_id = 0;
 	BYTE flags = 0;
 	uint16_t ec = 0;
@@ -1838,15 +2008,15 @@ static SECURITY_STATUS SEC_ENTRY kerberos_DecryptMessage(PCtxtHandle phContext,
 		return SEC_E_INVALID_TOKEN;
 
 	/* Read in header information */
-	header = sig_buffer->pvBuffer;
-	Data_Read_UINT16_BE(header, tok_id);
+	BYTE* header = sig_buffer->pvBuffer;
+	tok_id = winpr_Data_Get_UINT16_BE(header);
 	flags = header[2];
-	Data_Read_UINT16_BE((header + 4), ec);
-	Data_Read_UINT16_BE((header + 6), rrc);
-	Data_Read_UINT64_BE((header + 8), seq_no);
+	ec = winpr_Data_Get_UINT16_BE(&header[4]);
+	rrc = winpr_Data_Get_UINT16_BE(&header[6]);
+	seq_no = winpr_Data_Get_UINT64_BE(&header[8]);
 
 	/* Check that the header is valid */
-	if (tok_id != TOK_ID_WRAP || (BYTE)header[3] != 0xFF)
+	if ((tok_id != TOK_ID_WRAP) || (header[3] != 0xFF))
 		return SEC_E_INVALID_TOKEN;
 
 	if ((flags & FLAG_SENDER_IS_ACCEPTOR) == context->acceptor)
@@ -1882,18 +2052,21 @@ static SECURITY_STATUS SEC_ENTRY kerberos_DecryptMessage(PCtxtHandle phContext,
 		return SEC_E_INVALID_TOKEN;
 
 	/* Locate the parts of the message */
-	iov[0].data.data = header + 16 + rrc + ec;
+	iov[0].data.data = (char*)&header[16 + rrc + ec];
 	iov[1].data.data = data_buffer->pvBuffer;
-	iov[2].data.data = header + 16 + ec;
-	iov[3].data.data = iov[2].data.data + iov[2].data.length;
-	iov[4].data.data = iov[3].data.data + iov[3].data.length;
+	iov[2].data.data = (char*)&header[16 + ec];
+	char* data2 = iov[2].data.data;
+	iov[3].data.data = &data2[iov[2].data.length];
+
+	char* data3 = iov[3].data.data;
+	iov[4].data.data = &data3[iov[3].data.length];
 
 	if (krb_log_exec(krb5glue_decrypt_iov, creds->ctx, key, usage, iov, ARRAYSIZE(iov)))
 		return SEC_E_INTERNAL_ERROR;
 
 	/* Validate the encrypted header */
-	Data_Write_UINT16_BE(iov[2].data.data + 4, ec);
-	Data_Write_UINT16_BE(iov[2].data.data + 6, rrc);
+	winpr_Data_Write_UINT16_BE(iov[2].data.data + 4, ec);
+	winpr_Data_Write_UINT16_BE(iov[2].data.data + 6, rrc);
 	if (memcmp(iov[2].data.data, header, 16) != 0)
 		return SEC_E_MESSAGE_ALTERED;
 
@@ -1905,7 +2078,8 @@ static SECURITY_STATUS SEC_ENTRY kerberos_DecryptMessage(PCtxtHandle phContext,
 #endif
 }
 
-static SECURITY_STATUS SEC_ENTRY kerberos_MakeSignature(PCtxtHandle phContext, ULONG fQOP,
+static SECURITY_STATUS SEC_ENTRY kerberos_MakeSignature(PCtxtHandle phContext,
+                                                        WINPR_ATTR_UNUSED ULONG fQOP,
                                                         PSecBufferDesc pMessage, ULONG MessageSeqNo)
 {
 #ifdef WITH_KRB5
@@ -1914,7 +2088,6 @@ static SECURITY_STATUS SEC_ENTRY kerberos_MakeSignature(PCtxtHandle phContext, U
 	PSecBuffer data_buffer = NULL;
 	krb5glue_key key = NULL;
 	krb5_keyusage usage = 0;
-	char* header = NULL;
 	BYTE flags = 0;
 	krb5_crypto_iov iov[] = { { KRB5_CRYPTO_TYPE_DATA, { 0 } },
 		                      { KRB5_CRYPTO_TYPE_DATA, { 0 } },
@@ -1954,11 +2127,11 @@ static SECURITY_STATUS SEC_ENTRY kerberos_MakeSignature(PCtxtHandle phContext, U
 		return SEC_E_INSUFFICIENT_MEMORY;
 
 	/* Write the header */
-	header = sig_buffer->pvBuffer;
-	Data_Write_UINT16_BE(header, TOK_ID_MIC);
-	header[2] = flags;
+	char* header = sig_buffer->pvBuffer;
+	winpr_Data_Write_UINT16_BE(header, TOK_ID_MIC);
+	header[2] = WINPR_ASSERTING_INT_CAST(char, flags);
 	memset(header + 3, 0xFF, 5);
-	Data_Write_UINT64_BE(header + 8, (context->local_seq + MessageSeqNo));
+	winpr_Data_Write_UINT64_BE(header + 8, (context->local_seq + MessageSeqNo));
 
 	/* Set up the iov array */
 	iov[0].data.data = data_buffer->pvBuffer;
@@ -1978,14 +2151,14 @@ static SECURITY_STATUS SEC_ENTRY kerberos_MakeSignature(PCtxtHandle phContext, U
 
 static SECURITY_STATUS SEC_ENTRY kerberos_VerifySignature(PCtxtHandle phContext,
                                                           PSecBufferDesc pMessage,
-                                                          ULONG MessageSeqNo, ULONG* pfQOP)
+                                                          ULONG MessageSeqNo,
+                                                          WINPR_ATTR_UNUSED ULONG* pfQOP)
 {
 #ifdef WITH_KRB5
 	PSecBuffer sig_buffer = NULL;
 	PSecBuffer data_buffer = NULL;
 	krb5glue_key key = NULL;
 	krb5_keyusage usage = 0;
-	char* header = NULL;
 	BYTE flags = 0;
 	uint16_t tok_id = 0;
 	uint64_t seq_no = 0;
@@ -2009,10 +2182,10 @@ static SECURITY_STATUS SEC_ENTRY kerberos_VerifySignature(PCtxtHandle phContext,
 		return SEC_E_INVALID_TOKEN;
 
 	/* Read in header info */
-	header = sig_buffer->pvBuffer;
-	Data_Read_UINT16_BE(header, tok_id);
+	BYTE* header = sig_buffer->pvBuffer;
+	tok_id = winpr_Data_Get_UINT16_BE(header);
 	flags = header[2];
-	Data_Read_UINT64_BE((header + 8), seq_no);
+	seq_no = winpr_Data_Get_UINT64_BE((header + 8));
 
 	/* Validate header */
 	if (tok_id != TOK_ID_MIC)
@@ -2045,8 +2218,8 @@ static SECURITY_STATUS SEC_ENTRY kerberos_VerifySignature(PCtxtHandle phContext,
 
 	/* Set up the iov array */
 	iov[0].data.data = data_buffer->pvBuffer;
-	iov[1].data.data = header;
-	iov[2].data.data = header + 16;
+	iov[1].data.data = (char*)header;
+	iov[2].data.data = (char*)&header[16];
 
 	if (krb_log_exec(krb5glue_verify_checksum_iov, creds->ctx, key, usage, iov, ARRAYSIZE(iov),
 	                 &is_valid))
